pro convert_hcssfits_pacs, dir_in=dir_in, dir_out=dir_out, root=root, color=color, $
                           nsubs=nsubs, advanced_masks=advanced_masks, mask_transient=mask_transient


;;; conversion of Hipe level-1 fits files into Scanamorphos input structures, for PACS

;;; to be used one observation at a time: root name must include the obsid number
;;; convention Xra = -RA, Ydec = DEC

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; dir_in: directory where the Hipe file(s) can be found
;;; dir_out (optional): directory where the Scanamorphos input structures will be saved
;;;                     (= dir_in by default)
;;; root: root name of the Hipe file(s)
;;;       must allow to uniquely select the file associated with the obsid to process
;;;       (or the files if the observation was split into several parts)
;;; color: either "blue", "green", or "red"
;;;        (i.e. 70, 100, or 160 microns)
;;;        Please do not confuse blue and green observations
;;;        (same array but different filters and different parameters).
;;; nsubs (optional): Normally, one Hipe file contains a whole scan or several scans.
;;;                   If a scan is split between several Hipe files, set the "nsubs" parameter
;;;                   to the number of these files, and give them the same root name.
;;; advanced_masks (optional): Select this option if special masks were created for input
;;;                            to Scanamorphos with the HIPE script convertL1ToScanam.py
;;;                            (this is the recommended route).
;;;                            If not, bad samples should be set to NaN within Hipe
;;;                            before the level-1 data are saved to fits files.
;;; mask_transient (optional): Select this option if you want to mask transients
;;;                            after the calibration flash at the start of the obsid
;;;                            (only possible if advanced_masks is also selected).


;;; Before running Scanamorphos, include the full path name of dir_out
;;; in the first line of SCANAMORPHOS/scanlist_pacs
;;; and update the list of input structures, one per line.

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2010 H\'el\`ene Roussel and Marc Sauvage
;;; H.R.: roussel@iap.fr
;;;       Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France
;;; M.S.: marc.sauvage@cea.fr
;;;       CEA, Laboratoire AIM, Irfu/SAp, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; HISTORY
;;; MS: I try some modifications to also get the mask information
;;; MS: June 2 - 2006 The mask is modified in hipe so now we get 3 here



resolve_all, /continue_on_error, /quiet
crs = string(13b) + string(10b)

nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
if (not keyword_set(dir_out)) then dir_out = dir_in
nc = strlen(dir_out)
if strmid(dir_out, nc-1) ne '/' then dir_out = dir_out + '/'


if (not keyword_set(color)) then color = ''
if (color eq 'blue' or color eq 'green') then nb = 64L*32L $
   else if color eq 'red' then nb = 32L*16L else begin
      print, 'no array selected: exit'
      print, 'should be "blue", "green" or "red"'
      goto, end_label
   endelse

if (not keyword_set(nsubs)) then nsubs = 1


; MS the signal and status are indeed the first two extensions
; so they can be defined here.
exten_signal = 1
exten_status = 2   ;;; for time vector

; Ra and Dec will appear at different positions because the
; mask dataset is of variable size (more than one cube). 
; So it is better to look for their positions later.



name_search = root + '*.fits'
files = file_search(dir_in, name_search, /fold_case, /test_read, count=nf)
if nf ne nsubs then begin
   if nsubs eq 1 then print, crs + 'number of observation(s) found is not 1: exit' $
      else print, crs + 'number of scan parts found is not ' + chain(nsubs) + ': exit'
   print, 'Check directory and file names and permissions.'
   goto, end_label
endif else begin
   print, crs + 'observation found !' + crs
   for k = 0, nsubs - 1 do begin
      file = files(k)

      ; MS modifications
      ; The header lists the number of dataset stored in the
      ; extensions but one dataset can require more than one
      ; extension to be stored. In the header we have the
      ; relation between dataset number and extension number
      ; the masks are always one data set so Ra and Dec are the 13th
      ; and 14th dataset.  
      ; read the header of extension 0 (main header)
      main_hdr = headfits(file, exten=0)

      ; MS 02/08/10 - I now make it more robust: get all the extension names
      ; and find the location of those I'm interested in.

      ; this is the number of datasets
      ndataset = sxpar(main_hdr, 'dsets___')
      ; now get all their names.
      allnames_tab = strarr(ndataset)
      for iset = 0, ndataset - 1 do begin
         this_exten = sxpar(main_hdr, 'DS_'+chain(iset))
         allnames_tab(iset) = strtrim(sxpar(headfits(file, exten=this_exten), 'EXTNAME'), 2)
      endfor

      ; The datasets I need at this point are called 'Ra', 'Dec', and 'BooleanMasks'
      index = where(allnames_tab eq 'Ra', count)
      if (count ne 1) then begin
         print, 'ERROR -- Ra extension not found. This is fatal!'
         has_ra = 0
      endif else begin
         has_ra = 1
         ra_keyname = 'DS_' + chain(index(0))
         exten_ra = sxpar(main_hdr, ra_keyname)
      endelse
      index = where(allnames_tab eq 'Dec', count)
      if (count ne 1) then begin
         print, 'ERROR -- Dec extension not found. This is fatal!'
         has_dec = 0
      endif else begin
         has_dec = 1
         dec_keyname = 'DS_' + chain(index(0))
         exten_dec = sxpar(main_hdr, dec_keyname)
      endelse
      if (not(has_ra) or not(has_dec)) then goto, end_label
      index = where(allnames_tab eq 'BooleanMasks', count)
      if (count ne 1) then begin
         has_masks = 0
         print, 'ERROR - The boolean versions of the masks are not present'
         print, 'ERROR - No mask information will be propagated to Scanamorphos'
      endif else begin
         has_masks = 1
         ; use of the prefix 'last' is because this is usually the last one
         last_keyname = 'DS_' + chain(index(0))
         last_ext = sxpar(main_hdr, last_keyname)
      endelse

      if keyword_set(advanced_masks) then begin
         ; Now let's try to find the extension number of the 
         ; mask needed for scanamorphos (created in hipe)
         ; We basically use 3: SA_badpix is all the bad pixels
         ;                                  glitches, saturation...
         ;                     NonScience flags the preamble of the obs
         ;                     UNCLEANCHOP is the cal transient
         ;                                 not the best name.
         ; now last_ext is the extension number for the masks dataset
         ; so I get the header of that last exention
         mask_hdr = headfits(file, exten=last_ext)
         ; now find the number of datasets saved for the masks
         nsubdataset = sxpar(mask_hdr, 'dsets___')
         ; get the names of all the masks
         names_tab = strarr(nsubdataset)
         for iset = 0, nsubdataset - 1 do begin
            this_exten = sxpar(mask_hdr,'DS_'+chain(iset))
            names_tab(iset) = strtrim(sxpar(headfits(file, exten=this_exten),'EXTNAME'), 2)
         endfor

         ; bad pixels mask, called SA_badpix
         index = where(names_tab eq 'SA_badpix', count)
         if (count ne 1) then begin
            print, 'ERROR - The bad pixels mask is not found.'
            has_badpix = 0
         endif else begin
            has_badpix = 1
            this_exten = sxpar(mask_hdr, 'DS_'+chain(index(0)))
            maskBadPixel = read_hcss(file, this_exten)
         endelse

         ; NonScience mask, masking the preample to the actual scan
         index = where(names_tab eq 'NonScience', count)
         if (count ne 1) then begin
            print, 'ERROR - The NonScience mask is not found.'
            has_nonscience = 0
         endif else begin
            has_nonscience = 1
            this_exten = sxpar(mask_hdr, 'DS_'+chain(index(0)))
            maskNonScience = read_hcss(file, this_exten)
         endelse

         ; UNCLEANCHOP mask containing safety mask after calibration block(s)
         index = where(names_tab eq 'UNCLEANCHOP', count)
         if (count ne 1) then begin
            print, 'WARNING - The UncleanChop mask is not found.'
            has_uclchop = 0
         endif else begin
            has_uclchop = 1
            this_exten = sxpar(mask_hdr, 'DS_'+chain(index(0)))
            maskTransient = read_hcss(file, this_exten)
         endelse
         ;
         ; end of MS modifications
         ; The mask information is dealt with after reading the signal.
      endif

 
      scan_status = read_hcss(file, exten_status)
      scan_status = scan_status.__DATA__
      if k eq 0 then time_start = min(scan_status.finetime)
      t = 1.e-6 * (scan_status.finetime - time_start)
      nt = n_elements(t)
      if k eq 0 then time = t else time = [time, t]

      tags_status = strlowcase(tag_names(scan_status))
      w = where(tags_status eq 'repetition', cw)
      if cw eq 1 then begin
         ind = scan_status.repetition
         ind -= min(ind)
         nscans = max(ind) + 1
      endif else begin
         nscans = 1
         ind = intarr(nt)
      endelse
      if k eq 0 then ind_scan = ind else ind_scan = [ind_scan, ind]

      bbt = scan_status.bbtype
      w = where(bbt gt 0 and bbt lt 3282, cw)
      if cw ge 1 then print, 'excluding ' + chain(cw) + ' time step(s) on a total of ' + chain(nt) $
         + ': non-nominal frames'
      if k eq 0 then bbtype = bbt else bbtype = [bbtype, bbt]

      band = scan_status.band
      ; sometimes the first values are different because 
      ; we set the filter during the observation.
      ; using the middle point is better.
      print, ''
      filter = band(n_elements(band) / 2)
      print, 'wavelength band: ' + filter
      if ((strupcase(strmid(filter, 0, 1)) eq 'B' and color eq 'red') $
         or (strupcase(strmid(filter, 0, 1)) eq 'R' and color ne 'red')) then begin
         print, 'selected color does not match the array: exit'
         goto, end_label
      endif
      if (strupcase(filter) eq 'BS' and color eq 'green') then begin
         print, 'Changing the color from green to blue to match the data.'
         color = 'blue'
      endif
      if (strupcase(filter) eq 'BL' and color eq 'blue') then begin
         print, 'Changing the color from blue to green to match the data.'
         color = 'green'
      endif
      print, ''


      scan_signal = read_hcss(file, exten_signal)
      scan_signal = scan_signal.__DATA__
      sig = reform(scan_signal, nt, nb)


      if (keyword_set(advanced_masks) and has_masks) then begin
         ;;; use mask information here before concatenating the signal:
                                ; MS Modif: I mask in the signal all
                                ; the bad pixels as well as the non
                                ; science pixels. By default, the
                                ; transient after the calibration
                                ; block is not masked.
         ; bad pixels
         if (has_badpix) then begin
            scan_mask = reform(maskBadPixel.__DATA__, nt, nb)
            wz = where(scan_mask eq 1, cz)
            if cz ge 1 then sig(wz) = !values.f_nan
         endif
         ; Non science data
         if (has_nonscience) then begin
            scan_mask = reform(maskNonScience.__DATA__, nt, nb)
            wz = where(scan_mask eq 1, cz)
            if cz ge 1 then sig(wz) = !values.f_nan
         endif
         ;;; transients after calibration flash:
         if (keyword_set(mask_transient) and has_uclchop) then begin
            scan_mask = reform(maskTransient.__DATA__, nt, nb)
            wz = where(scan_mask eq 1, cz)
            if cz ge 1 then sig(wz) = !values.f_nan
         endif
      endif

      if k eq 0 then signal = sig else signal = [signal, sig]


      scan_ra_deg = read_hcss(file, exten_ra)
      scan_ra_deg = scan_ra_deg.__DATA__
      Xra = -reform(scan_ra_deg, nt, nb)
      scan_dec_deg = read_hcss(file, exten_dec)
      scan_dec_deg = scan_dec_deg.__DATA__
      Ydec = reform(scan_dec_deg, nt, nb)
      if k eq 0 then begin
         Xra_bolos = Xra
         Ydec_bolos = Ydec
      endif else begin
         Xra_bolos = [Xra_bolos, Xra]
         Ydec_bolos = [Ydec_bolos, Ydec]
      endelse
   endfor


   ;;; exclude time jumps:

   nt = n_elements(time)
   diff_t = [1., time(1:nt-1) - time(0:nt-2)]
   w1 = where(diff_t le 0., cw)
   if cw ge 1 then begin
      flag = intarr(nt)
      for i = 0, cw - 1 do begin
         w2 = w1(i) + where(time(w1(i):*) gt time(w1(i)-1), c2)
         if c2 ge 1 then flag(w1(i):w2(0)-1) = 1 else flag(w1(i):*) = 1
      endfor

      wkeep = where(flag lt 1, nt, ncomplement=cz)
      print, 'excluding ' + chain(cz) + ' time step(s) (absurd time vector)'
      time = time(wkeep)
      ind_scan = ind_scan(wkeep)
      bbtype = bbtype(wkeep)
      signal = signal(wkeep, *)
      Xra_bolos = Xra_bolos(wkeep, *)
      Ydec_bolos = Ydec_bolos(wkeep, *)
   endif


   ;;; mask invalid bolometers:

   w = where(finite(signal) eq 0 or finite(Xra_bolos) eq 0 or finite(Ydec_bolos) eq 0, cw)
   if cw ge 1 then begin
      wb = w / nt
      histo = histogram(wb, min=0., max=nb-1)
      wkeep = where(histo lt 9 * nt / 10, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' bolometers'
         signal(*, wz) = !values.f_nan
         print, ''
      endif
   endif



   ;;; loop on scans:

   for s = 0, nscans - 1 do begin

      ;;; to select scan s and exclude spurious calibration data:
      ws = where(ind_scan eq s and (bbtype lt 1 or bbtype gt 3281))
      time_s = time(ws)

      ;;; create structure:
      if color eq 'blue' then begin
         signal_psw = signal(ws, *)
         Xra_bolos_psw = Xra_bolos(ws, *)
         Ydec_bolos_psw = Ydec_bolos(ws, *)
         scan = create_struct('time_s', time_s, 'signal_psw', signal_psw, $
            'Xra_bolos_psw', Xra_bolos_psw, 'Ydec_bolos_psw', Ydec_bolos_psw, $
            'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
      endif else if color eq 'green' then begin
         signal_pmw = signal(ws, *)
         Xra_bolos_pmw = Xra_bolos(ws, *)
         Ydec_bolos_pmw = Ydec_bolos(ws, *)
         scan = create_struct('time_s', time_s, 'signal_pmw', signal_pmw, $
            'Xra_bolos_pmw', Xra_bolos_pmw, 'Ydec_bolos_pmw', Ydec_bolos_pmw, $
            'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
      endif else if color eq 'red' then begin
         signal_plw = signal(ws, *)
         Xra_bolos_plw = Xra_bolos(ws, *)
         Ydec_bolos_plw = Ydec_bolos(ws, *)
         scan = create_struct('time_s', time_s, 'signal_plw', signal_plw, $
            'Xra_bolos_plw', Xra_bolos_plw, 'Ydec_bolos_plw', Ydec_bolos_plw, $
            'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
      endif

      file_out = 'stru_' + root + '_scan' + chain(s+1) + '.xdr'
      save, filename=dir_out+file_out, scan, /xdr
      print, 'saved file "' + file_out + '"'
   endfor
   print, ''

endelse


end_label:

end
